; keybdf84.asm - Morse Code keyboard keyer add-on using the 16F84 PIC
;
; updated 07/15/2020, Chuck Olson, WB9KZY - per Henry, ZS5HR, the PIC pinout
;	section of the comments had an error, the keyer output and keyer dit
;	input pins were swapped.  This was corrected.
;
; updated 11/27/2001, Chuck Olson, WB9KZY - bug found - leading edge of the
;       keyed output has a .3 ms glitch (0=>1=>0) when the PK-Basic or Tick
;       keyer chips are used.  A much smaller glitch occurs with the K8 keyer
;       chip.  No glitch is present with the PK-3 keyer chip.
;       problem solved by disabling the pullups in the init routine, now
;       they are enabled only during a rdbyte call.
;
;
;
; 2/17/2000     Chuck Olson, WB9KZY
;               Jackson Harbor Press
;               http://jacksonharbor.home.att.net
;               jacksonharbor@att.net
;
;
;
;       What does it do?
;
;       This program uses a 16f84 along with a keyer and an IBM keyboard
;       to allow keyboard sending of Morse code - using Island Memory tech.
;       See the file: keybdrea.dme for more information on the program.
;
;
;
;       Possible Enhancements:
;
;       1) add usage of the keyboard LEDs to signal buffer "fullness"?
;       2) better editing in memory recording
;       3) use control/shift keys for more and/or smaller memories?
;       4) use an output pin connected to a piezo speaker for a keyboard 
;               "click" when keys are pressed
;       5) use an output pin to turn on LEDs to signal the buffer "fullness"
;               (easier than item 1??)
;       6) add a manual pause to the message play or buffer play - press a
;               key and the play will pause until it's pressed again
;               this would allow manual paddle input to the keyer or keyboard
;       7) add embedded pause - designate one key (and character code) as
;               a manual pause byte - this could be inserted into the messages
;               during record and would allow the message play to pause
;               for manual input from the keyboard or keyer paddle
;       8) add a embeddable or special key contest counter of some kind?
;       9) add some kind of practice mode
;      10) add sidetone, element timing to get rid of the external keyer
;      11) add a powerdown mode to turn off the keyboard after a certain
;               period of inactivity - need a PIC input pin to serve as
;               an on/off button?
;      12) add recognition of the E0/E1 prefix to allow usage of the other
;               keys (keypad, arrows, et cetera).
;
;
;
;       PIC pinout 
;       (also, see the file:  keybdf84.pdf for the project schematic)
;PIC
;pin    name    function/connection
;---    ----    -----------------------------------------------
;  1    ra2     no connect
;  2    ra3     no connect
;  3    ra4     no connect
;  4    MCLR    pulled up to +5V with a 10K resistor
;  5    VSS     Ground 
;  6    rb0     keyboard clock line
;  7    rb1     keyboard data line
;  8    rb2     24LC16 external eeprom clock line, scl, pin 6 (optional)
;  9    rb3     24LC16 external eeprom data line, sdb, pin 5 (optional)
; 10    rb4     keyer output
; 11    rb5     keyer dah input
; 12    rb6     keyer dit input
; 13    rb7     keyer switch input (optional)
; 14    VDD     +5V
; 15    osc2    connected to 4 MHz xtal & 22? pf cap
; 16    osc1    connected to 4 MHz xtal & 22? pf cap
; 17    ra0     no connect
; 18    ra1     no connect



        LIST P=16F84, R=DEC 
        INCLUDE "p16F84.inc"
        __FUSES _PWRTE_OFF & _CP_OFF & _WDT_OFF & _XT_OSC
; powerup timer off - still not sure why
; copy protect is off
; watchdog timer is off 
; the oscillator uses an external 4 Mhx crystal or resonator



; memory locations
temp    equ     H'0C'   ;all purpose temp byte - used in halfcyc
char    equ     H'0D'   ;holder for character data - used in playchr, strtbit
size    equ     H'0E'   ;char size counter - used in playchr, strtbit
third   equ     H'0F'   ;holder for number of dit length spaces => ditsp
temp2   equ     H'10'   ;holder for element length => ditsp, deboun
flags   equ     H'11'   ;flag bits such as kill, start, stop, parity
eeaddr  equ     H'12'   ;internal eeprom address holder => reccal, playcal, 
                        ; => eepread, wreep and external eeprom routines
ditlen  equ     H'13'   ;dit length in 1 ms increments => outdit, ditsp
savew   equ     H'14'   ;temporary location to save w during an interrupt
saves   equ     H'15'   ;temporary location to save STATUS during an interrupt
ebit    equ     H'16'   ;8 bit counter location for keyboard => rupt, intinit
inbyte  equ     H'17'   ;data loaded by the keyboard => rupt, intinit
recv    equ     H'18'   ;receive buffer address => rupt, initbuf, mainlup
xmit    equ     H'19'   ;transmit buffer addr => initbuf, mainlup, recm3, reccal
savef   equ     H'1A'   ;temporary location to save FSR during an interrupt
temp3   equ     H'1B'   ;used to test the buffered character b7 in mainlup
makectr equ     H'1C'   ;make counter - keeps track of the number of times a
                        ;key code is sent before a break => rupt, mainlup
eom     equ     H'1D'   ;end of memory counter => reccal
;location 1E-22 are defined/used in the external eeprom routines (end of pgm)
bufstrt equ     H'23'   ;start of 45 character buffer
bufstop equ     H'4F'   ;end of buffer

; constants - also see individual routines

; bit definitions
carry   equ     0       ;bit 0 of STATUS is the carry bit
zero    equ     2       ;bit 2 of STATUS is the zero bit
break   equ     0       ;bit 0 of flags: preceeding character a break?, 1=yes
kill    equ     1       ;bit 1 of flags: kill bit, 1=kill message play
start   equ     2       ;bit 2 of flags: start bit complete?, 1=completed
parity  equ     3       ;bit 3 of flags: parity bit complete?, 1=completed
chrdone equ     4       ;bit 4 of flags: character complete?, 1=completed
record  equ     5       ;bit 5 of flags: enter record mode?, 1=yes
err     equ     6       ;bit 6 of flags: error flag for ext eeprom routines
clock   equ     0       ;bit 0 of PORTB is the keyboard clock bit
kbdata  equ     1       ;bit 1 of PORTB is the keyboard data bit
;rem'd the next two lines as they are defined in the external eeprom routines
;SCL     equ     2       ;bit 2 of PORTB is the external eeprom clock line
;SDA     equ     3       ;bit 3 of PORTB is the external eeprom data line
owt     equ     4       ;bit 4 of PORTB is the keyer output
dah     equ     5       ;bit 5 of PORTB is the dah output
dit     equ     6       ;bit 6 of PORTB is the dit output
switch  equ     7       ;bit 7 of PORTB is the switch output
pu      equ     7       ;bit 7 of the option register controls the pullups



        ORG     0x00
begin   goto    init    ;startup routine
        


        ORG     0x04
        goto    rupt    ;branch over the tables



; decoder - a table to decode the scan codes from the PC-AT keyboard to
;               Morse code equivalents - all non Morse keys are zero'd
; Morse code character storage structure from Dave Gwillym, KB2TQX
; right justified storage of the character:
;                                       dit = 0
;                                       dah = 1
; the leftmost 1 is a start bit - the byte is left shifted until a
;       1 is seen, then the dits and dahs are shifted out for a
;       total of 8 shifts - this holds most normal characters.
decoder addwf   PCL,1   ;compute the jump value
;                                dec hex key name
;                                --- --  -----------------------------------
        retlw   0               ;    00
        retlw   0xF9            ;    01  f9 key - clever coding, eh??
        retlw   0               ;    02  should be f7 key but >>isn't<<
        retlw   0xF5            ;    03  f5 key
        retlw   0xF3            ;    04  f3 key
        retlw   0xF1            ;    05  f1 key
        retlw   0xF2            ;    06  f2 key
        retlw   0xFC            ;    07  f12 key
        retlw   0               ;    08
        retlw   0xFA            ;    09  f10 key
        retlw   0xF8            ; 10 0a  f8 key
        retlw   0xF6            ; 11 0b  f6 key
        retlw   0xF4            ; 12 0c  f4 key & backspace
        retlw   0               ; 13 0d  tab key
        retlw   0               ; 14 0e  tilde (` key (upper left - next to 1 key))
        retlw   0               ; 15 0f  end & 1 (keypad)

        retlw   0               ; 16 10
        retlw   0               ; 17 11  left alt & 4(keypad)
        retlw   0               ; 18 12  left shift key & home & 7(keypad)
        retlw   0               ; 19 13
        retlw   0               ; 20 14  pause & left ctrl & right ctrl ??
        retlw   b'00011101'     ; 21 15  Q
        retlw   b'00101111'     ; 22 16  1
        retlw   0               ; 23 17  right & left alt & right arrow
        retlw   0               ; 24 18
        retlw   0               ; 25 19
        retlw   b'00011100'     ; 26 1a  Z
        retlw   b'00001000'     ; 27 1b  S
        retlw   b'00000101'     ; 28 1c  A
        retlw   b'00001011'     ; 29 1d  W
        retlw   b'00100111'     ; 30 1e  2
        retlw   0               ; 31 1f  left windows

        retlw   0               ; 32 20
        retlw   b'00011010'     ; 33 21  C
        retlw   b'00011001'     ; 34 22  X
        retlw   b'00001100'     ; 35 23  D
        retlw   b'00000010'     ; 36 24  E
        retlw   b'00100001'     ; 37 25  4
        retlw   b'00100011'     ; 38 26  3
        retlw   0               ; 39 27  right windows
        retlw   0               ; 40 28
        retlw   b'01111111'     ; 41 29  space - arb. set space to 7Fh
        retlw   b'00010001'     ; 42 2a  V
        retlw   b'00010010'     ; 43 2b  F
        retlw   b'00000011'     ; 44 2c  T
        retlw   b'00001010'     ; 45 2d  R
        retlw   b'00100000'     ; 46 2e  5
        retlw   0               ; 47 2f  box and arrow key (W95)

        retlw   0               ; 48 30
        retlw   b'00000110'     ; 49 31  N
        retlw   b'00011000'     ; 50 32  B
        retlw   b'00010000'     ; 51 33  H
        retlw   b'00001110'     ; 52 34  G
        retlw   b'00011011'     ; 53 35  Y
        retlw   b'00110000'     ; 54 36  6
        retlw   0               ; 55 37
        retlw   0               ; 56 38
        retlw   0               ; 57 39
        retlw   b'00000111'     ; 58 3a  M
        retlw   b'00010111'     ; 59 3b  J
        retlw   b'00001001'     ; 60 3c  U
        retlw   b'00111000'     ; 61 3d  7
        retlw   b'00111100'     ; 62 3e  8
        retlw   0               ; 63 3f

        retlw   0               ; 64 40
        retlw   b'01110011'     ; 65 41  ,
        retlw   b'00001101'     ; 66 42  K
        retlw   b'00000100'     ; 67 43  I
        retlw   b'00001111'     ; 68 44  O
        retlw   b'00111111'     ; 69 45  0 (numeric top row)
        retlw   b'00111110'     ; 70 46  9
        retlw   0               ; 71 47
        retlw   0               ; 72 48
        retlw   b'01010101'     ; 73 49  .(above space bar)
        retlw   b'01001100'     ; 74 4a  this is the ?/ key
        retlw   b'00010100'     ; 75 4b  L
        retlw   b'01000101'     ; 76 4c  ;  => SK prosign
        retlw   b'00010110'     ; 77 4d  P
        retlw   b'00101010'     ; 78 4e  - (numeric top row) => AR prosign
        retlw   0               ; 79 4f

        retlw   0               ; 80 50
        retlw   0               ; 81 51
        retlw   b'11000101'     ; 82 52  single quote: '  => BK prosign
        retlw   0               ; 83 53
        retlw   b'00101000'     ; 84 54  [  => AS prosign
        retlw   b'00110001'     ; 85 55  = (send as BT)
        retlw   0               ; 86 56
        retlw   0               ; 87 57
        retlw   0               ; 88 58  caps lock
        retlw   0               ; 89 59  right shift key
        retlw   0               ; 90 5a  enter(mid) & enter(keypad)
        retlw   b'00110110'     ; 91 5b  ]  => KN prosign
        retlw   0               ; 92 5c
        retlw   b'00110010'     ; 93 5d  \(back slash)
        retlw   0               ; 94 5e
        retlw   0               ; 95 5f

        retlw   0               ; 96 60
        retlw   0               ; 97 61
        retlw   0               ; 98 62
        retlw   0               ; 99 63
        retlw   0               ;100 64
        retlw   0               ;101 65
        retlw   0               ;102 66  back space key
        retlw   0               ;103 67
        retlw   0               ;104 68
        retlw   0               ;105 69
        retlw   0               ;106 6a
        retlw   0               ;107 6b 
        retlw   0               ;108 6c             
        retlw   0               ;109 6d
        retlw   0               ;110 6e
        retlw   0               ;111 6f

        retlw   0               ;112 70  0 (keypad) & insert
        retlw   0               ;113 71  . (keypad) & delete
        retlw   0               ;114 72  down arrow & 2(keypad)
        retlw   0               ;115 73  5(keypad)
        retlw   0               ;116 74  6(keypad) & right arrow & back slash (\)
        retlw   0               ;117 75  8(keypad) & up arrow
        retlw   0xF0            ;118 76  esc
        retlw   0               ;119 77  num lock key
        retlw   0xFB            ;120 78  f11
        retlw   0               ;121 79  +
        retlw   0               ;122 7a  3(keypad) & page down
        retlw   0               ;123 7b  -(keypad)
        retlw   0               ;124 7c  print screen & *(keypad)
        retlw   0               ;125 7d  9(keypad) & page up (middle top)
        retlw   0               ;126 7e  scroll lock key
        retlw   0               ;127 7f
        retlw   0               ;128 80
        retlw   0               ;129 81
        retlw   0               ;130 82
        retlw   0xF7            ;131 83  f7



; jtab - a jump table for decoding the function keys & esc
jtab    movlw   0xF0            ;adjust the jump pointer by F0
        subwf   temp3,0         ;do the subtraction leaving the result in w
        addwf   PCL,1           ;compute the jump value
        return                  ;f0 = escape key - decoded elsewhere
        goto    playcal         ;f1 = play 1st internal eeprom message
        goto    playm2          ;f2 = play 2nd internal eeprom message
        goto    play3           ;f3 = play 3rd external eeprom message
        goto    play4           ;f4 = play 4th external eeprom message
        goto    play5           ;f5 = play 5th external eeprom message
        goto    play6           ;f6 = play 6th external eeprom message
        goto    play7           ;f7 = play 7th external eeprom message
        goto    play8           ;f8 = play 8th external eeprom message
        goto    play9           ;f9 = play 9th external eeprom message
        goto    play10          ;f10 = play 10th external eeprom message
        return                  ;f11 - decoded elsewhere
        return                  ;f12 - decoded elsewhere



; jtab2 - a jump table for decoding the function keys & esc for recording
jtab2   movlw   0xF0            ;adjust the jump pointer by F0
        subwf   temp3,0         ;do the subtraction leaving the result in w
        addwf   PCL,1           ;compute the jump value
        return                  ;f0 = escape key - decoded elsewhere
        goto    reccal          ;f1 = record 1st internal eeprom message
        goto    recm2           ;f2 = record 2nd internal eeprom message
        goto    recm3           ;f3 = record 3rd external eeprom message
        goto    recm4           ;f4 = record 4th external eeprom message
        goto    recm5           ;f5 = record 5th external eeprom message
        goto    recm6           ;f6 = record 6th external eeprom message
        goto    recm7           ;f7 = record 7th external eeprom message
        goto    recm8           ;f8 = record 8th external eeprom message
        goto    recm9           ;f9 = record 9th external eeprom message
        goto    recm10          ;f10 = record 10th external eeprom message
        return                  ;f11 - decoded elsewhere
        return                  ;f12 - decoded elsewhere

  

;  reccal - recm2 - internal eeprom recording routines
reccal  movlw   b'00101111'     ;load a Morse 1
        call    playchr         ;play the 1
        clrf    eeaddr          ;clear the internal eeprom address byte
        movlw   0x1F            ;last recordable location in memory 1
        movwf   eom             ;store it in eom for later testing
        goto    strtrec         ;start recording
recm2   movlw   b'00100111'     ;load a Morse 2
        call    playchr         ;play the 2
        movlw   0x20            ;load up the start of message 2 (20 hex)
        movwf   eeaddr          ;save it in the internal eeprom address byte
        movlw   0x3F            ;last recordable location in memory 2
        movwf   eom             ;store it in eom for later testing
strtrec btfsc   flags,kill      ;has the kill flag been set?
        goto    killrec         ;yes, kill the recording
        ;look at the decoded character in the buffer
        movf    xmit,0          ;load up the xmit pointer
        movwf   FSR             ;put xmit pointer into the indirect address
        movf    INDF,0          ;get the contents of the buffer
        movwf   temp3           ;store the byte just for a little bit
        movf    temp3,0         ;reload the byte from the buffer 
                                ;WHY? to set/reset the zero bit in STATUS
        btfsc   STATUS,zero     ;is this byte a zero?
        goto    strtrec         ;yes, keep looping for now until it's not zero
        btfsc   temp3,7         ;is this a function key?
        goto    skipl2          ;yes, skip the character play and recording
        ;no, first see if the character is special (space = 7fh)
        sublw   0x7F            ;compare it to a space
        btfsc   STATUS,zero     ;is it a space?
        goto    wait            ;yes, send a wdsp2 instead of the 7f charact
        movf    temp3,0         ;reload the character
        call    playchr         ;play the character        
skiplay movf    temp3,0         ;load the character into w
        movwf   EEDATA          ;put it in EEDATA to prepare for wreep
        call    wreep           ;write the character into eeprom
        incf    eeaddr,1        ;increment the eeprom address
        ;a test for memory boundry - prevent writing over the other memory
        movf    eom,0           ;load up the last possible memory location
        subwf   eeaddr,w        ; is the xmit pointer 
        btfsc   STATUS,zero     ;       equal to the last possible memory loc?
        goto    killrec         ;yes, stop recording and zero the last loc
skipl2  clrf    INDF            ;zero the buffer
        incf    xmit,1          ;increment the xmit pointer
        movlw   bufstop+1       ;load up the last possible buffer location + 1
        subwf   xmit,w          ; is the xmit pointer 
        btfss   STATUS,zero     ;       equal to the last possible buffer loc?
        goto    strtrec         ;no, loop back for another character
        movlw   bufstrt         ;yes, reset the xmit pointer 
        movwf   xmit            ;                       to the bufstrt
        goto    strtrec         ;loop back for another character
wait    call    wdsp2           ;wait for a word space
        goto    skiplay         ;branch back around the character play
; 
;  service the kill bit by zeroing the buffer, recording a zero and return
killrec call    initbuf         ;initialize the buffer
        bcf     flags,kill      ;clear the kill bit
        clrf    EEDATA          ;clear the eeprom data byte
        call    wreep           ;write the character into eeprom
        return                  ;end of reccal routine



; rupt - the interrupt routine - go here when the keyboard clock goes low
;       this interrupt service routine will:
;       1) get the key code
;       2) decide what it is (key break, func key or char to be buffered)
;       3) fill the circular buffer
;               two pointers are used
;               recv points to the last location filled with data
;               xmit points to the last location played
;
; per an email from Steve Weber, (KD1JV):
;
; at powerup, zero the buffer - both pointers to start
;
; when you get a char from the keyboard (via interrupt?)
;       look at byte pointed to by recv - if <> 0 then you've overflowed
;               if =0 then it's OK, put the character into the byte
;
; when you send a character, write a zero into it's location
;       
; Note: for entended keys (insert,home,page up, page down, end, delete and
;       the arrow) the keyboard will send: E0 F0 (break) and then the key code
;
rupt    ;save the status and w registers before proceeding with the int svc
        movwf   savew           ;save the w register
        swapf   STATUS,0        ;load the status register into w
        movwf   saves           ;save the status register
        swapf   FSR,0           ;load the indirect address register
        movwf   savef           ;save the indirect address register
        ;also, disable the interrupts? - don't bother - done automagically 
        btfss   flags,start     ;is this the start bit?
        goto    startb          ;yes, set the start complete bit & exit
        ;no, proceed with the character
        btfss   flags,chrdone   ;is the character xfer complete bit set?
        goto    nextbit         ;no, get the next bit of the character
        ;yes, see if the parity bit was sent
        btfss   flags,parity    ;is this the parity bit?
        goto    paritb          ;yes, set the parity complete bit
        ;since all other flag bits are set, this must be the stop bit
        incf    makectr,1       ;increment the make counter
        ;so first, see if the char is a break code (F0)
        movlw   0xF0            ;load up a release code        
        subwf   inbyte,0        ;compare the input byte with F0 (release code)
        btfss   STATUS,zero     ;is this a release code?
        goto    dcode           ;no, go and decode the character
        bsf     flags,break     ;yes, set the break code flag
        goto    cintin          ;bail out
        ;if it's not a break, decode the character
dcode   ;check for key codes above 0x83 (f7)
        movlw   0x84            ;is this code greater than the higest key (f7)?
        subwf   inbyte,0        ;compare the input byte with the ack byte
        btfsc   STATUS,zero     ;are they the same?
        goto    cbreak          ;yes, bail out, don't do a look up
        btfsc   STATUS,carry    ;is the key code less than x84?
        goto    cbreak          ;no, bail out, don't do a lool up
        ;check for the 78h (dit - f11) key code
        movlw   0x78            ;load up an F11 key code
        subwf   inbyte,0        ;compare the input byte with F0 (release code)
        btfss   STATUS,zero     ;is this a release code?
        goto    f12tst          ;no, go and check if it's f12
        btfsc   flags,break     ;has the break code been sent?
        goto    ditoff          ;yes, turn off the dit
        bsf     STATUS,5        ;set the bank bit to 1
        bcf     TRISB,dit       ; set dit to an output
        bcf     STATUS,5        ;reset the bank bit to 0
        bcf     PORTB,dit       ;set the dit low
        goto    cbreak          ;skip the buffering
ditoff  bsf     STATUS,5        ;set the bank bit to 1
        bsf     TRISB,dit       ; set bit 5 (dit) to an input
        bcf     STATUS,5        ;reset the bank bit to 0
        goto    cbreak          ;skip the buffering
        ;check for the 07h (dah - f12) key code
f12tst  movlw   0x07            ;load up an F12 key code
        subwf   inbyte,0        ;compare the input byte with F0 (release code)
        btfss   STATUS,zero     ;is this a release code?
        goto    scrlock         ;no, proceed to the scroll lock test
        btfsc   flags,break     ;has the break code been sent?
        goto    dahoff          ;yes, turn off the dah
        bsf     STATUS,5        ;set the bank bit to 1
        bcf     TRISB,dah       ; set bit 5 (dah) to an output
        bcf     STATUS,5        ;reset the bank bit to 0
        bcf     PORTB,dah       ;set the dah low
        goto    cbreak          ;skip the buffering
dahoff  bsf     STATUS,5        ;set the bank bit to 1
        bsf     TRISB,dah       ; set bit 5 (dah) to an input
        bcf     STATUS,5        ;reset the bank bit to 0
        goto    cbreak          ;skip the buffering
        ;check for the 7Eh (switch - scroll lock) key code
scrlock movlw   0x7E            ;load up the scroll lock key code
        subwf   inbyte,0        ;compare the input byte with F0 (release code)
        btfss   STATUS,zero     ;is this a release code?
        goto    lookup          ;no, proceed to the lookup
        btfsc   flags,break     ;has the break code been sent?
        goto    switoff         ;yes, turn off the switch
        bsf     STATUS,5        ;set the bank bit to 1
        bcf     TRISB,switch    ; set switch to an output
        bcf     STATUS,5        ;reset the bank bit to 0
        bcf     PORTB,switch    ;set the switch low
        goto    cbreak          ;skip the buffering
switoff bsf     STATUS,5        ;set the bank bit to 1
        bsf     TRISB,switch    ;set switch to an input
        bcf     STATUS,5        ;reset the bank bit to 0
        goto    cbreak          ;skip the buffering
lookup  movf    inbyte,0        ;load up w for
        call    decoder         ;       a decode of the character
        ;next, evaluate the characer - is it a 0? - if yes, ignore it
        movwf   inbyte          ;save the decoded char in inbyte
        movf    inbyte,0        ;test the decoded character
        btfsc   STATUS,zero     ;is the decoded char is a zero?
        goto    cbreak          ;it is, ignore it.
        ;check the make counter - is it over 32? (should we record something?)
        btfsc   makectr,5       ;is bit 5 of the counter a 1? (makectr >= 32 ?)
        bsf     flags,record    ;yes, set the record bit in flags
        ;is it after a break? if no, ignore it
        btfss   flags,break     ;was the preceeding character a break?
        goto    cintin          ;no, bail out
        ;does the user want to kill the message play? (check for escape)
        movlw   0xF0            ;is this an esc?
        subwf   inbyte,0        ;compare the input byte with an extended char
        btfsc   STATUS,zero     ;are they the same?
        goto    setkill         ;yes, go and set the kill bit
        ;it's a Morse character, save the decoded character in the buffer
        movf    recv,0          ;load up the recv pointer
        movwf   FSR             ;put recv pointer into the indirect address
        movf    INDF,0          ;get the contents of the buffer
        btfss   STATUS,zero     ;is this byte a zero?
        goto    cbreak          ;no, the buffer is full, just bail for now
        ;yes, store the character
        movf    inbyte,0        ;load up the current character
        movwf   INDF            ;save it in the 1 char buffer
        incf    recv,1          ;increment the recv pointer
        movlw   bufstop+1       ;load up the last possible buffer location + 1
        subwf   recv,w          ; is the recv pointer 
        btfss   STATUS,zero     ;       equal to the last possible buffer loc?
        goto    cbreak          ;no, exit the interrupt routine
        movlw   bufstrt         ;yes, reset the recv pointer 
        movwf   recv            ;                       to the bufstrt
cbreak  bcf     flags,break     ;clear the break flag
        clrf    makectr         ;clear the make counter after a break or ?
cintin  call    intinit         ;initialize the various buffers, counters
        goto    exint           ;bail out of the interrup routine
startb  bsf     flags,start     ;set the start complete bit
        goto    exint           ;go and exit the interrupt routine
nextbit btfsc   PORTB,kbdata    ;check the keyboard data bit   
        goto    setcar          ;it's a one, set the carry
        bcf     STATUS,carry    ;it's a zero, clear the carry
        goto    rotat           ;jump over the carry set
setcar  bsf     STATUS,carry    ;set the carry
rotat   rrf     inbyte,1        ;shift the data bit into the input byte
        decfsz  ebit,1          ;decrement the 8 bit counter - is it 0?
        goto    exint           ;no, go and exit the interrupt routine 
        bsf     flags,chrdone   ;yes, set the character done flag
        goto    exint           ;exit the interrupt routine
paritb  bsf     flags,parity    ;set the parity complete bit
        goto    exint           ;go and exit the interrupt routine
setkill bsf     flags,kill      ;set the kill flag
        goto    cbreak          ;bail out
exint   ;restore the status and w registers
        swapf   savef,0         ;load the saved FSR register into w
        movwf   FSR             ;put it back into FSR
        swapf   saves,0         ;load the saved status register into w
        movwf   STATUS          ;put it back into STATUS
        ;note that the swapf instruction doesn't affect STATUS bits
        swapf   savew,1         ;swap the nibbles of the saved w register
        swapf   savew,0         ;swap them again into the w register
        bcf     INTCON,INTF     ;re-enable port b 0 int by clearing flag bit
        bsf     INTCON,INTE     ;enable port b 0 interrupt                
        retfie  ;return from the interrupt - end of rupt routine



; outdit - a routine to send a dit
outdit  clrf    ditlen          ;clear the dit length counter
        bsf     STATUS,5        ;set the bank bit to 1
        bcf     TRISB,dit       ; set bit 6 (dit) to an output        
        bcf     STATUS,5        ;reset the bank bit to 0
        bcf     PORTB,dit       ;simulate a dit paddle closure
        call    halfcyc         ;wait a little
ditlup  btfss   PORTB,owt       ;has the keyer output changed? (gone high)
        goto    ditlup          ;no, loop back until the keyer output changes
        bsf     STATUS,5        ;set the bank bit to 1
        bsf     TRISB,dit       ; set bit 6 (dit) to an input
        bcf     STATUS,5        ;reset the bank bit to 0
incditl incf    ditlen,1        ;increment the dit length counter
        call    halfcyc         ;waste 500 usec
        call    halfcyc         ; and waste another 500 usec for 1 ms total
        btfss   PORTB,dit       ;has the dit been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        btfss   PORTB,dah       ;has the dah been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        btfsc   PORTB,owt       ;has the keyer output gone low?
        goto    incditl         ;no, go back and waste another 1 ms & count it
        return                  ;end of the outdit routine
        


; outdah - a routine to send a dah
outdah  bsf     STATUS,5        ;set the bank bit to 1
        bcf     TRISB,dah       ; set bit 5 (dah) to an output
        bcf     STATUS,5        ;reset the bank bit to 0
        bcf     PORTB,dah       ;set the dah low
dahlup  call    halfcyc         ;wait a little
        btfss   PORTB,owt       ;has the keyer output changed? (gone high)
        goto    dahlup          ;no, loop back until the keyer output changes
        bsf     STATUS,5        ;set the bank bit to 1
        bsf     TRISB,dah       ; set bit 5 (dah) to an input
        bcf     STATUS,5        ;reset the bank bit to 0
dahlup2 call    halfcyc         ;wait a little - 1/2 ms
        call    halfcyc         ;wait a little more - another 1/2 ms
        btfss   PORTB,dit       ;has the dit been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        btfss   PORTB,dah       ;has the dah been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        btfsc   PORTB,owt       ;has the keyer output gone low?
        goto    dahlup2         ;no, go back and wait until it does
        return                  ;end of the outdah routine



; ditsp - a routine to wait 1 dit space
; wdsp - a routine to wait 6 dit spaces + 1 from last dit or dah = 7 total 
; wdsp2 - a routine to wait 4 dit spaces + 3 from last char = 7 total 
; chsp - a routine to wait 2 dit spaces + 1 from last dit or dah = 3 total
ditsp   movlw   0x2             ;load up 1 dit space (add one due to loop contstruct)
        goto    loadet          ;branch in
wdsp    movlw   0x8             ;load up 7 dit spaces
        goto    loadet          ;branch in
wdsp2   movlw   0x5             ;load up 4 dit space (an additional delay after chsp)
        goto    loadet          ;branch into the routine with an alterante
chsp    movlw   0x4             ;load up 3 dit spaces
loadet  movwf   third           ;save it in third - a temp reg
decet   decfsz  third,1         ;decrement the dit space counter - is it 0 ?
        goto    ditsp2          ;no, start waiting
        goto    dun2            ;yes, end the dit
ditsp2  ;movlw   d'80'          ;load a 15 wpm dit length
        movf    ditlen,0        ;load up the dit length
        movwf   temp2           ;store it in the second temp location
loop6   decfsz  temp2,1         ;decrement the dit length counter - is it 0 ?
        goto    nobeep2         ;no, start waiting
        goto    decet           ;yes, check the dit space counter
nobeep2 call    halfcyc         ;wait a half audio cycle
        call    halfcyc         ;wait a half audio cycle
        btfss   PORTB,dit       ;has the dit been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        btfss   PORTB,dah       ;has the dah been pressed?
        bsf     flags,kill      ;yes, set the kill bit
        goto    loop6           ;go back for another cycle?
dun2    return                  ;end of a bunch of dit space routines



; halfcyc - a routine to wait 1/2 of a sidetone cycle - in this case
;       that is 500 usec or 1/2 ms => 1000 Hz tone
;       at PIC clock = 4 MHz, 500 usec = 500 PIC instruction cycles
; delay = 2 (call) + 2 (return) + 5  (routine) + ((freq x 3) - 1)  loop delay
freq    equ     d'164'  ; 9 + (164 x 3) - 1 = 500
halfcyc movlw   freq            ;1   1 load up the 1/2 cycle delay constant
        movwf   temp            ;1   2 store it in the temporary location
        goto    waste2          ;2   4 waste 2 cycles
waste2  nop                     ;1   5 waste a cycle
loopy   decfsz  temp,1          ;1     decrement the temp location - is it 0 ?
        goto    loopy           ;2     no, keep decrementing
        ;yes, return to the calling routine
        return                  ;2 500 end of half cycle routine



; strtbit - a routine to rotate a char left until a 1 appears
strtbit movlw   0x8             ;load up the maximum char size + 1
        movwf   size            ;save it temp
strbit2 decf    size,1          ;decrement the element counter
rotit   rlf     char,1          ;rotate the msb into carry
        btfss   STATUS,0        ;test the carry - 0=placeholder, 1=start bit
        goto    strbit2         ;it's a just a placeholder (0), keep shifting
        return                  ;you're done - end of the startbit routine            

        
; playchr - a routine to play a character stored in w with Morse code
playchr movwf   char            ;save it in char
getdig3 call    strtbit         ;rotate char left until the start bit comes up
shout3  rlf     char,1          ;rotate the msb into carry
        btfsc   STATUS,0        ;test the carry - 0=dit, 1=dah
        goto    dahsnd3         ;it's a one, send a dah
        ; send a dit
        call    outdit          ;send a dit
        goto    sizdec          ;go check the size variable
dahsnd3 call    outdah          ;send the dah
sizdec  decfsz  size,1          ;decrement the char size counter, is it 0?        
        goto    shout3          ;no, go back for the next bit
        ;yes, send a char space and then kill this thing
        call    chsp            ;send a character space
rpcend  return                  ;end of playchr routine



; init - the initialization power up entry point of the program
init    bsf     STATUS,5        ;set the bank bit to 1
        movlw   b'11110011'     ; all inputs on PORTB, except bits 2,3
        movwf   TRISB           ; set the tri-state register
        movlw   b'00000000'     ;PORTA all outs 
        movwf   TRISA           ; set the tri-state register
; 11/27/2001, disabled the pullups on port b due to the glitch problem
        movlw   b'10000000'     ; PORTB p/u disabled, int on falling edge of rb0
        movwf   OPTION_REG      ; set the option register
        bcf     STATUS,5        ;set the bank bit to 0
        bcf     PORTB,dit       ;turn on the dit
        bcf     PORTB,dah       ;       and dah outputs
        bcf     PORTB,switch    ;  and switch outputs        
        movlw   d'60'           ;20 wpm dit length
        movwf   ditlen          ;load up a startup default dit space of 20 wpm
        clrf    makectr         ;clear the make counter
        call    initbuf         ;initialize the buffer
        clrf    flags           ;clear all the flags
        call    intinit         ;setup for the interrupt
        bcf     INTCON,INTF     ;clear the port b 0 interrupt flag
        bsf     INTCON,GIE      ;enable interrupts
        bsf     INTCON,INTE     ;enable port b 0 interrupt                



; mainlup - the main loop in the program - executes continuously after init
;       flushes out the circular buffer
mainlup btfsc   flags,kill      ;has the kill flag been set?
        goto    killmsg         ;yes, kill the message play
        btfsc   flags,record    ;has the record flag been set?
        goto    recit           ;yes, enter the record mode
        ;look at the decoded character in the buffer
        movf    xmit,0          ;load up the xmit pointer
        movwf   FSR             ;put xmit pointer into the indirect address
        movf    INDF,0          ;get the contents of the buffer
        movwf   temp3           ;store the byte just for a little bit
        ;check for the BK prosign - the only Morse char with bit 7 high
        sublw   0xC5            ;compare it to a BK
        btfsc   STATUS,zero     ;is it a BK?
        goto    ztest           ;yes, just play the character
        btfss   temp3,7         ;is bit 7 high (80-FF?)?
        goto    ztest           ;no, skip the function key stuff
funkey  call    jtab            ;go and use the jump table to call the
                                ;correct subroutine
        goto    ovrplay         ;skip over the zero test, etc.
ztest   movf    temp3,0         ;reload the byte from the buffer
        btfsc   STATUS,zero     ;is this byte a zero?
        goto    mainlup         ;yes, keep looping for now until it's not zero
        ;no, first see if the character is special (space = 7fh)
        sublw   0x7F            ;compare it to a space
        btfss   STATUS,zero     ;is it a space?
        goto    playit          ;no, just play the character
        call    wdsp2           ;yes, go and send a word space
        goto    ovrplay         ;skip over the character play
playit  movf    INDF,0          ;reload the character
        call    playchr         ;play the character        
ovrplay btfsc   flags,kill      ;has the kill flag been set?
        goto    killmsg         ;yes, kill the message play
        clrf    INDF            ;zero the buffer
        incf    xmit,1          ;increment the xmit pointer
        movlw   bufstop+1       ;load up the last possible buffer location + 1
        subwf   xmit,w          ; is the xmit pointer 
        btfss   STATUS,zero     ;       equal to the last possible buffer loc?
        goto    mainlup         ;no, loop back for another character
        movlw   bufstrt         ;yes, reset the xmit pointer 
        movwf   xmit            ;                       to the bufstrt
        goto    mainlup         ;loop back for another character
; 
;  service the kill bit by zeroing the buffer and then resuming mainlup
killmsg call    initbuf         ;initialize the buffer
bufclr  bcf     flags,kill      ;clear the kill bit
        goto    mainlup         ;go back to the main loop
;
;  record one of the message memories
recit   bcf     flags,record    ;clear the record bit
        clrf    makectr         ;clear the make counter
        movlw   b'01001100'     ;load up a question mark
        call    playchr         ;play it
        ;check here for bufstrt underflow
        movlw   bufstrt         ;load up bufstrt for the comparison
        subwf   recv,0          ;compare the present recv with bufstrt
        btfss   STATUS,zero     ;are they equal?
        goto    decrecv         ;no, just decrement recv
        movlw   bufstop+1       ;yes, adjust recv to the top of the buffer + 1
        ;get the last key code in the buffer which should be the msg key
decrecv decf    recv,0          ;decrement and load the recv pointer
        movwf   FSR             ;put recv pointer into the indirect address
        movf    INDF,0          ;get the contents of the buffer
        movwf   temp3           ;save it for the jump table computation
        call    initbuf         ;kill the buffer
        call    jtab2           ;enter the recording jump table
        goto    mainlup         ;last statement in mainlup routine



; initbuf - clear the buffer to all zero and set the pointers
initbuf movlw   bufstrt         ;load up the start of the buffer
        movwf   FSR             ;set the buffer pointer
agn     movlw   d'0'            ;load a zero into w
        movwf   INDF            ;zero the buffer location
        incf    FSR,1           ;increment the memory pointer
        movlw   bufstop         ;lets look at the element counter
        subwf   FSR,W           ;is the element counter 
        btfss   STATUS,Z        ;        less than the message end address?
        goto    agn             ;yes, they are unequal - keep going
        movlw   bufstrt         ;load up the starting address of the buffer
        movwf   xmit            ;set the transmit buffer pointer
        movwf   recv            ;set the receive buffer pointer
        return                  ;end of initbuf routine



; intinit - a routine to initialize registers for a keyboard interrupt        
;       zero the flags, reset the counters, etc.
intinit bcf     flags,start     ;clear the start,
        ;bcf     flags,break     ;       break,
        bcf     flags,chrdone   ;   character done
        bcf     flags,parity    ;       and carry bits
        movlw   8               ;prepare for an 8 bit character xfer
        movwf   ebit            ;save it in the ebit down counter
        clrf    inbyte          ;clear the input byte
        return                  ;end of intinit routine



; deboun - wait for 20 ms
deboun  movlw   d'40'           ;load up a counter
        movwf   temp2           ;store it in temp2
dbounc  call    halfcyc         ;wait 500 usec
        decfsz  temp2,1         ;decrement the loop count - skip at 0
        goto    dbounc          ;count not zero, loop back
        return                  ;end of debounce routine



; playcal - a routine to play the EEPROM messages - playcal plays the
;       message starting at address 0 - playm2 plays the message starting
;       at address 20 - both are up to 32 characters long
playcal clrf    eeaddr          ;store the start address of the callsign
        goto    ldele2          ;skip the second entry point into this routine
playm2  movlw   0x20            ;start of the callsign stored in eeprom is 3
        movwf   eeaddr          ;store the start address of the callsign
ldele2  btfsc   flags,kill      ;has the kill flag been set?
        goto    endplay         ;yes, go and kill this message play
        ;no, continue with the message play
        call    eepread         ;get the byte from internal eeprom
        btfsc   STATUS,zero     ;is this the end of the message (0)??
        goto    endplay         ;yes, end the playcal routine
        ;no, test for a word space
        sublw   0x7F            ;a 7F hex is a space
        btfss   STATUS,zero     ;is this a space??
        goto    ovrspac         ;yes, go and send a wdsp2, then loop back
        call    wdsp2           ;send a word space        
        goto    branchi         ;branch back into the routine
ovrspac movf    EEDATA,0        ;no, reload the character
mark2   call    playchr         ;play the character
branchi incf    eeaddr,1        ;increment the eeprom address
        goto    ldele2          ;get the next element
endplay return                  ;yes, end of the playcal routine



; eepread - a routine to increment the eeprom address pointer and then
;       read the eeprom data at that new address into w
eepread ;incf    eeaddr,1        ;increment the eeprom address pointer, eeaddr
skipinc movf    eeaddr,0        ;load eeaddr into w        
        movwf   EEADR           ;save it in EEADR
        bsf     STATUS,5        ;set the bank bit to 1 for EECON1 set
        bsf     EECON1,RD       ;setup internal EEPROM for read mode
        bcf     STATUS,5        ;reset the bank bit to 0
        movf    EEDATA,0        ;load up w with the EEPROM data byte
        return                  ;end of the eepread routine



;    wreep - a routine to write the data in EEDATA into the address EEADR
; write sequence from the 16F84 spec - for the internal 16f84 eeprom.
; Why use the eeaddr memory location to hold the internal eeprom address
;       rather than using EEADR?
; Because EEADR appears to be reset to 0 after each write to the internal
;       eeprom, so it can't be used as an address pointer for the record2 
;       routine - this doesn't happen with the read from the internal eeprom.
wreep   movf    eeaddr,0        ;load up the internal eeprom address
        movwf   EEADR           ;store is in the EEPROM address buffer
        bsf     STATUS,5        ;set the bank bit to 1
        bsf     EECON1,WREN     ;enable write
        movlw   H'55'           ;load up a 55h per the 16f84 spec
        movwf   EECON2          ;write the 55h 
        movlw   H'AA'           ;load up AAh per the 16f84 spec
        movwf   EECON2          ;write the AAh 
        bsf     EECON1,WR       ;set the WR bit and thus begin writing to eeprom
        bcf     EECON1,WREN     ;disable write
checkit btfss   EECON1,EEIF     ;see if the last write is complete
        goto    checkit         ;no, check it again
        bcf     EECON1,EEIF     ;clear the write complete interrupt flag bit
        bcf     STATUS,5        ;reset the bank bit to 0
        return                  ;end of wreep routine



; play3 - play10 - routines to play the 8 messages in external eeprom
; press and release f3 to play message 3
; press and release f4 to play message 4
; press and release f5 to play message 5
; press and release f6 to play message 6
; press and release f7 to play message 7
; press and release f8 to play message 8
; press and release f9 to play message 9
; press and release f10 to play message 10 
play3   ;start playing the EEPROM contents
        movlw   0               ;init the start of the message buffer page
        goto    ldele3          ;skip the other entry points
play4   ;start playing the EEPROM contents
        movlw   1               ;page of message 4
        goto    ldele3          ;skip the other entry points
play5   ;start playing the EEPROM contents
        movlw   2               ;page of message 5
        goto    ldele3          ;skip the other entry points
play6   ;start playing the EEPROM contents
        movlw   3               ;page of message 6
        goto    ldele3          ;skip the other entry points
play7   ;start playing the EEPROM contents
        movlw   4               ;page of message 7
        goto    ldele3          ;skip the other entry points
play8   ;start playing the EEPROM contents
        movlw   5               ;page of message 8
        goto    ldele3          ;skip the other entry points
play9   ;start playing the EEPROM contents
        movlw   6               ;page of message 9
        goto    ldele3          ;skip the other entry points
play10  ;start playing the EEPROM contents
        movlw   7               ;page of message 10
ldele3  movwf   eepage          ;init the start of the message buffer page
        clrf    ADDR            ;start of the message stored in eeprom is zero
ldele3a btfsc   flags,kill      ;has the kill flag been set?
        goto    endply3         ;yes, go and kill this message play
        bsf     STATUS,RP0      ;set the bank bit to 1
        bcf     OPTION_REG,pu   ;turn on the pullups
        bcf     STATUS,RP0      ;set the bank bit to 0
        call    RDBYTE          ;get a byte
        bsf     STATUS,RP0      ;set the bank bit to 1
        bsf     OPTION_REG,pu   ;turn off the pullups
        bcf     STATUS,RP0      ;set the bank bit to 0
        movf    DATAI,0         ;load up w with the EEPROM data byte
        btfsc   STATUS,zero     ;is this the end of the message (0)??
        goto    endply3         ;yes, end the play
        ;no, first see if the character is special (space = 7fh)
        sublw   0x7F            ;compare it to a space
        btfss   STATUS,zero     ;is it a space?
        goto    playit2         ;no, just play the character
        call    wdsp2           ;yes, go and send a word space
        goto    ovplay          ;skip over the character play
playit2 movf    DATAI,0         ;reload up w with the EEPROM data byte
        call    playchr         ;send the character
        ;add an element and/or page counter check here ?
ovplay  incfsz  ADDR,1          ;increment the element counter
        goto    lt256a          ;it's not zero, leave the page alone
        goto    endply3         ;you've overflowed the page, bail out
lt256a  goto    ldele3a         ;yes, get the next element
endply3 return                  ;end of the play3-10 routines



;  recm3 - recm10 - routines to record messages into external EEPROM
; The data is stored in the 24LC16 external eeprom which requires totally 
; different routines than the internal eeprom of the 16f84.
; press and hold f3 (wait for q mark) + 3 to record message 3
;       and so on for f4 - f10
recm3   movlw   b'00100011'     ;load up a 3 for memory 3
        call    playchr         ;play it
        movlw   0               ;page 0 for memory 3 (256 bytes long)
        goto    startr          ;branch over the other entry points
recm4   movlw   b'00100001'     ;load up a 4
        call    playchr         ;play it
        movlw   1               ;page 1
        goto    startr          ;branch over the other entry points
recm5   movlw   b'00100000'     ;load up a 5
        call    playchr         ;play it
        movlw   2               ;page 2
        goto    startr          ;branch over the other entry points
recm6   movlw   b'00110000'     ;load up a 6
        call    playchr         ;play it
        movlw   3               ;page 3
        goto    startr          ;branch over the other entry points
recm7   movlw   b'00111000'     ;load up a 7
        call    playchr         ;play it
        movlw   4               ;page 4
        goto    startr          ;branch over the other entry points
recm8   movlw   b'00111100'     ;load up a 8
        call    playchr         ;play it
        movlw   5               ;page 5
        goto    startr          ;branch over the other entry points
recm9   movlw   b'00111110'     ;load up a 9
        call    playchr         ;play it
        movlw   6               ;page 6
        goto    startr          ;branch over the other entry points
recm10  movlw   b'00101111'     ;load up a 1
        call    playchr         ;play it
        movlw   b'00111111'     ;load up a 0
        call    playchr         ;play it
        movlw   7               ;page 7
startr  movwf   eepage          ;init the start of the message buffer page
        clrf    ADDR            ;load up the start of the message buffer
startra btfsc   flags,kill      ;has the kill flag been set?
        goto    killr           ;yes, kill the recording
        ;look at the decoded character in the buffer
        movf    xmit,0          ;load up the xmit pointer
        movwf   FSR             ;put xmit pointer into the indirect address
        movf    INDF,0          ;get the contents of the buffer
        movwf   temp3           ;store the byte just for a little bit
        btfsc   STATUS,zero     ;is this byte a zero?
        goto    startra         ;yes, keep looping for now until it's not zero
        btfsc   temp3,7         ;is this a function key?
        goto    skipr2          ;yes, skip the character play and recording
        ;no, first see if the character is special (space = 7fh)
        movlw   0x7F            ;load up a space character for comparision
        subwf   temp3,0         ;compare it to a space
        btfsc   STATUS,zero     ;is it a space?
        goto    vvait           ;yes, go and call wdsp2
        movf    temp3,0         ;no, reload the character
        call    playchr         ;play the character        
skplay  movf    temp3,0         ;load the character into w
        movwf   DATAO           ;move it to the buffer
        call    WRBYTE          ;write the byte
        incfsz  ADDR,1          ;increment the element counter
        goto    skipr2          ;it's not zero, leave the page alone
        goto    killr           ;yes, stop recording and zero the last loc
skipr2  clrf    INDF            ;zero the buffer
        incf    xmit,1          ;increment the xmit pointer
        movlw   bufstop+1       ;load up the last possible buffer location + 1
        subwf   xmit,0          ; is the xmit pointer 
        btfss   STATUS,zero     ;       equal to the last possible buffer loc?
        goto    startra         ;no, loop back for another character
        movlw   bufstrt         ;yes, reset the xmit pointer 
        movwf   xmit            ;                       to the bufstrt
        goto    startra         ;loop back for another character
vvait   ;this call was added to provide needed delay for the eeprom write
        call    wdsp2           ;do a word space if the space bar has been pressed
        goto    skplay          ;branch back over the character play
killr   ;service kill bit by zeroing the buffer, recording a zero and return
        call    initbuf         ;initialize the buffer
        bcf     flags,kill      ;clear the kill bit
        clrf    DATAO           ;load a 0 for the end of message byte
        call    WRBYTE          ;write the EOM byte to the eeprom
        return                  ;end of the recm3 - recm10 routines
       
        

; the following code is from Microchip AN515
;*****************************************************************************
;**  Two wire/I2C Bus READ/WRITE Sample Routines of Microchip's
;**  24Cxx / 85Cxx serial CMOS EEPROM interfacing to a 
;**  PIC16C54 8-bit CMOS single chip microcomputer
;**  Revsied Version 2.0 (4/2/92).
;**
;**     Part use = PIC16C54-XT/JW                                           
;**  Note:  1) All timings are based on a reference crystal frequency of 2MHz
;**            which is equivalent to an instruction cycle time of 2 usec.
;**         2) Address and literal values are read in octal unless otherwise
;**            specified.
;
;
;       Program:          I2CBUS.ASM 
;       Revision Date:   
;                         12-12-95      Compatibility with MPASMWIN 1.30
;
;********************************************************************** 
;
;*****************************************************************************
;
;-----------------------------------------------------------------------------
;       Files Assignment
;-----------------------------------------------------------------------------
;
FLAG    EQU     flags           ; Common flag bits register
EEPROM  EQU     EEDATA          ; Bit buffer
ADDR    EQU     eeaddr          ; Address register
DATAI   EQU     eom             ; Stored data input register
DATAO   EQU     H'22'           ; Stored data output register
; substituted the memory location, eepage, instead of SLAVE
eepage  equ     H'21'           ; 256 byte page in memory
TXBUF   EQU     H'20'           ; TX buffer
RXBUF   EQU     H'1F'           ; RX buffer
COUNT   EQU     H'1E'           ; Bit counter
;
;as far as I can tell, TIMER0 and TIMER1 aren't used, so they were deleted               
;ERCODE  equ     H''     ; error code byte - is it ever read? - no, delete it
;
;
;-----------------------------------------------------------------------------
;                     Bit Assignments
;-----------------------------------------------------------------------------
;
; Status bits
;
Z       EQU     2
C       EQU     0
;
; FLAG Bits
;
;ERR1    EQU     0               ; Error flag
ERR1    EQU     err             ; Error flag
;
; Instruction Destination Bits
;
F       EQU     1
W       EQU     0
;
; EEPROM Bits
;
DI      EQU     3               ; EEPROM input
DO      EQU     2               ; EEPROM output
;
; I2C Device Bits
;
SDA     EQU     3               ; RB3, data in/out
SCL     EQU     2               ; RB2, serial clock
;
; Constants
;
;END FILES/BITS EQUATE
; PAGE
;
;-----------------------------------------------------------------------------
;       Two wire/I2C - CPU communication error status table and subroutine
;-----------------------------------------------------------------------------
;  input  :     W-reg   = error code
;  output :     ERCODE  = error code
;           FLAG(ERR1) = 1
;
;         code          error status mode
;       -------         ------------------------------------------------------
;           1   :       SCL locked low by device (bus is still busy)
;           2   :       SDA locked low by device (bus is still busy)
;           3   :       No acknowledge from device (no handshake)
;           4   :       SDA bus not released for master to generate STOP bit
;-----------------------------------------------------------------------------
;
;Subroutine to identify the status of the serial clock (SCL) and serial data
;(SDA) condition according to the error status table. Codes generated are
;useful for bus/device diagnosis.
;
ERR
        BTFSS   FLAG,ERR1       ; Remain as first error encountered
        ;MOVWF   ERCODE          ; Save error code
        nop                     ;decided not to use ERCODE as it's never read
        BSF     FLAG,ERR1       ; Set error flag
        RETLW   0
;
;-----------------------------------------------------------------------------
;       START bus communication routine
;-----------------------------------------------------------------------------
;       input   : none
;       output  : initialize bus communication
;-----------------------------------------------------------------------------
;
;Generate START bit (SCL is high while SDA goes from high to low transition)
;and check status of the serial clock.
BSTART
;        MOVLW   B'00000000'     ; Put SCL, SDA line in output state
;        MOVLW   B'11110011'     ; Put SCL, SDA line in output state
        bsf     STATUS,5        ;set the bank bit to 1
        ;movwf   TRISB           ; set the tri-state register
        bcf     TRISB,SCL       ;set the scl to an output
        bcf     TRISB,SDA       ;set the sda to an output
        bcf     STATUS,5        ;clear the bank bit to 0
;***************************************************************************
        bsf     PORTB,SDA          ;make sure sda is high
;***************************************************************************
        BSF     PORTB,SCL          ; Set clock high
        MOVLW   1               ; Ready error status code 1
        BTFSS   PORTB,SCL          ; Locked?
        CALL    ERR             ; SCL locked low by device
        BCF     PORTB,SDA          ; SDA goes low during SCL high
        NOP                     ; Timing adjustment
        NOP
        NOP
        BCF     PORTB,SCL          ; Start clock train
        RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       STOP bus communication routine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       Bus communication, STOP condition
;-----------------------------------------------------------------------------
;
;Generate STOP bit (SDA goes from low to high during SCL high state)
;and check bus conditions.
;
BSTOP
;****************************************************************************
        ;MOVLW   B'00000000'     ; Put SCL, SDA line in output state
        ;MOVLW   B'11110011'     ; Put SCL, SDA line in output state
        bsf     STATUS,5        ;set the bank bit to 1
        ;movwf   TRISB           ; set the tri-state register
        bcf     TRISB,SCL       ;set the scl to an output
        bcf     TRISB,SDA       ;set the sda to an output
        bcf     STATUS,5        ;clear the bank bit to 0
;****************************************************************************
        BCF     PORTB,SDA          ; Return SDA to low
        BSF     PORTB,SCL          ; Set SCL high
        nop
        nop
        nop
        MOVLW   1               ; Ready error code 1
        BTFSS   PORTB,SCL          ; High?
        CALL    ERR             ; No, SCL locked low by device
        BSF     PORTB,SDA          ; SDA goes from low to high during SCL high
        MOVLW   4               ; Ready error code 4
        BTFSS   PORTB,SDA          ; High?
        CALL    ERR             ; No, SDA bus not release for STOP
        RETLW   0
;
;END SUB
;
;-----------------------------------------------------------------------------
;       Serial data send from PIC to serial EEPROM, bit-by-bit subroutine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       To (DI) of serial EEPROM device
;-----------------------------------------------------------------------------
;
BITIN
        ;MOVLW   B'00001000'     ; Force SDA line as input
        bsf     STATUS,5        ;set the bank bit to 1
        ;movwf   TRISB           ; set the tri-state register
        bcf     TRISB,SCL       ;set the scl to an output
        bsf     TRISB,SDA       ;set the sda to an input
        bcf     STATUS,5        ;clear the bank bit to 0
        BSF     PORTB,SDA          ; Set SDA for input
        BCF     EEPROM,DI
        BSF     PORTB,SCL          ; Clock high
        MOVLW   1
        BTFSC   PORTB,SCL          ; Skip if SCL  is high
        GOTO    BIT1
        BTFSS   FLAG,ERR1       ; Remain as first error encountered
        ;MOVWF   ERCODE          ; Save error code
        nop                     ;decided not to use ERCODE as it's never read
        BSF     FLAG,ERR1       ; Set error flag
BIT1
        BTFSC   PORTB,SDA          ; Read SDA pin
        BSF     EEPROM,DI       ; DI = 1
        NOP                     ; Delay
        BCF     PORTB,SCL          ; Return SCL to low     
        RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       Serial data receive from serial EEPROM to PIC, bit-by-bit subroutine
;-----------------------------------------------------------------------------
;       Input   :       EEPROM file
;       Output  :       From (DO) of serial EEPROM device to PIC
;-----------------------------------------------------------------------------
;
BITOUT
        ;MOVLW   B'00000000'     ; Set SDA, SCL as outputs
        ;MOVLW   B'11110011'     ; Put SCL, SDA line in output state
        bsf     STATUS,5        ;set the bank bit to 1
        ;movwf   TRISB           ; set the tri-state register
        bcf     TRISB,SCL       ;set the scl to an output
        bcf     TRISB,SDA       ;set the sda to an output
        bcf     STATUS,5        ;clear the bank bit to 0
        BTFSS   EEPROM,DO
        GOTO    BIT0
        BSF     PORTB,SDA          ; Output bit 0
        MOVLW   2
        BTFSC   PORTB,SDA          ; Check for error code 2
        GOTO    CLK1
        BTFSS   FLAG,ERR1       ; Remain as first error encountered
        ;MOVWF   ERCODE          ; Save error code
        nop                     ;decided not to use ERCODE as it's never read
        BSF     FLAG,ERR1       ; Set error flag
        GOTO    CLK1            ; SDA locked low by device
;
BIT0
        BCF     PORTB,SDA          ; Output bit 0
        NOP                     ; Delay
        NOP
        NOP
CLK1
        BSF     PORTB,SCL
        MOVLW   1               ; Error code 1
        BTFSC   PORTB,SCL          ; SCL locked low?
        GOTO    BIT2            ; No.
        BTFSS   FLAG,ERR1       ; Yes.
        ;MOVWF   ERCODE          ; Save error code
        nop                     ;decided not to use ERCODE as it's never read
        BSF     FLAG,ERR1       ; Set error flag
BIT2
        NOP
        NOP
        BCF     PORTB,SCL          ; Return SCL to low
        RETLW   0
;
;END SUB
; PAGE
;
;
;-----------------------------------------------------------------------------
;       RECEIVE         DATA    subroutine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       RXBUF = Receive 8-bit data
;-----------------------------------------------------------------------------
;
RX
        MOVLW   .8              ; 8 bits of data
        MOVWF   COUNT
        CLRF    RXBUF
;
RXLP
        RLF     RXBUF, F        ; Shift data to buffer
        SKPC
        BCF     RXBUF,0         ; carry ---> f(0)
        SKPNC
        BSF     RXBUF,0
        CALL    BITIN
        BTFSC   EEPROM,DI
        BSF     RXBUF,0         ; Input bit =1
        DECFSZ  COUNT, F        ; 8 bits?
        GOTO    RXLP
        BSF     EEPROM,DO       ; Set acknowledge bit = 1
        CALL    BITOUT          ; to STOP further input
        RETLW   0
;
;END SUB
;
;-----------------------------------------------------------------------------
;       TRANSMIT        DATA    subroutine
;-----------------------------------------------------------------------------
;       Input   :       TXBUF
;       Output  :       Data X'mitted to EEPROM device
;-----------------------------------------------------------------------------
;
TX
        MOVLW   .8
        MOVWF   COUNT
;
TXLP
        BCF     EEPROM,DO       ; Shift data bit out.
        BTFSC   TXBUF,7         ; If shifted bit = 0, data bit = 0
        BSF     EEPROM,DO       ; Otherwise data bit = 1
        CALL    BITOUT          ; Serial data out
        RLF     TXBUF, F        ; Rotate TXBUF left
        SKPC                    ; f(6) ---> f(7)
        BCF     TXBUF,0         ; f(7) ---> carry
        SKPNC                   ; carry ---> f(0)
        BSF     TXBUF,0
        DECFSZ  COUNT, F        ; 8 bits done?
        GOTO    TXLP            ; No.
        CALL    BITIN           ; Read acknowledge bit
        MOVLW   3
        BTFSC   EEPROM,DI       ; Check for acknowledgement
        CALL    ERR             ; No acknowledge from device
        RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       BYTE-WRITE, write one byte to EEPROM device
;-----------------------------------------------------------------------------
;       Input   :       DATAO= data to be written
;                       ADDR    = destination address
;                       eepage  = 256 byte page address (1010xxx0)
;       Output  :       Data written to EEPROM device
;-----------------------------------------------------------------------------
;
WRBYTE
        rlf     eepage,0        ;mult page by 2 and load into w
        andlw   b'00001110'     ;mask out bits 0,4-7
        iorlw   b'10100000'     ;"add" in the rest of the bits
        MOVWF   TXBUF           ; to TX buffer
        CALL    BSTART          ; Generate START bit
        CALL    TX              ; Output page address
        MOVF    ADDR,W          ; Get WORD address
        MOVWF   TXBUF           ; into buffer
        CALL    TX              ; Output WORD address
        MOVF    DATAO,W         ; Move DATA 
        MOVWF   TXBUF           ; into buffer
        CALL    TX              ; Output DATA and detect acknowledgement
        CALL    BSTOP           ; Generate STOP bit
        return
;
;
;
;-----------------------------------------------------------------------------
;       BYTE-READ, read one byte from serial EEPROM device
;-----------------------------------------------------------------------------
;       Input   :       ADDR    = source address
;                       eepage  = 256 byte page address (1010xxx0)
;       Output  :       DATAI   = data read from serial EEPROM
;-----------------------------------------------------------------------------
;
RDBYTE
        rlf     eepage,0        ;mult page by 2 and load into w
        andlw   b'00001110'     ;mask out bits 0,4-7
        iorlw   b'10100000'     ;"add" in the rest of the bits
        MOVWF   TXBUF           ; into buffer (R/W = 0)
        CALL    BSTART          ; Generate START bit
        CALL    TX              ; Output page address. Check ACK.
        MOVF    ADDR,W          ; Get WORD address
        MOVWF   TXBUF
        CALL    TX              ; Output WORD address. Check ACK.
        CALL    BSTART          ; START READ (if only one device is
        rlf     eepage,0        ;mult page by 2 and load into w
        andlw   b'00001110'     ;mask out bits 0,4-7
        iorlw   b'10100000'     ;"add" in the rest of the bits
        MOVWF   TXBUF
        BSF     TXBUF,0         ; Specify READ mode (R/W = 1)
        CALL    TX              ; Output page address
        CALL    RX              ; READ in data and acknowledge
        CALL    BSTOP           ; Generate STOP bit
        MOVF    RXBUF,W         ; Save data from buffer
        MOVWF   DATAI           ; to DATAI file.
        return          ;end of rdbyte routine
      

;setup the 16F84 eeprom for two messages
; the messages are stored in the same format as the decoder table
;       a 1 is a dah, a 0 is a dit, starting on the left there is a one 
;       used as a start bit, for example an A is stored as:
;
;            retlw   b'00000101'   OR   retlw   0x05  ;(hex)
;                           |||
;                           ||+-- dah
;                           |+-- dit
;                           +-- start bit
;
        org     0x2100  ;starting address for 16F84 internal eeprom
        de      0x0F            ;load an O
        de      0x0D            ;load a K
        de      0       ;end of message byte
        
        org     0x2120  ;starting address for 16F84 internal eeprom
        de      0x0B            ;load a W
        de      0x18            ;load a B
        de      0x3E            ;load a 9
        de      0x0D            ;load a K
        de      0x1C            ;load a Z
        de      0x1B            ;load a Y
        de      0       ;end of message byte        


        END     ;end of the keybdf84.asm program

;       copyright 2000 by Charles J. Olson 
